﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using System.ComponentModel;
using ThinkGeo.Core;
using ThinkGeo.UI.Wpf;
using System.Windows.Media.Imaging;

namespace ThinkGeoPerformanceApp
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        private MyBaseFeature _mySymbolFeature;
        private MyBaseFeature _myShapeFeature;

        private LayerOverlay _featureOverlay;

        public static GeoImage _state0GeoImage;
        public static GeoImage _state1GeoImage;
        public static GeoImage _state2GeoImage;
        public static GeoImage _state3GeoImage;
        public static GeoImage _stateNoGeoImage;

        public static MapView _gmap;
        public float _labelXOffsetInPixels = 0.0F;
        public float _labelYOffsetInPixels = 0.0F;

        private List<string> _labelLocations = new List<string>();
        private string _selectedLabelLocation = TextPlacement.Lower.ToString();

        public event PropertyChangedEventHandler PropertyChanged;

        public string SelectedLabelLocation
        {
            get
            {
                return _selectedLabelLocation;
            }
            set
            {
                _selectedLabelLocation = value;
                TextPlacement labelLocation = (TextPlacement)Enum.Parse(typeof(TextPlacement), value);
                //_mySymbolFeature.LabelLocation = labelLocation;
                _myShapeFeature.LabelLocation = labelLocation;
                wpfMap1.Refresh(_featureOverlay);
                OnPropertyChanged("SelectedLabelLocation");
            }
        }

        public List<string> LabelLocations
        {
            get
            {
                return _labelLocations;
            }
        }

        public float LabelXOffsetInPixels
        {
            get
            {
                return _labelXOffsetInPixels;
            }
            set
            {
                _labelXOffsetInPixels = value;
                _mySymbolFeature.LabelXOffsetInPixels = _labelXOffsetInPixels;
                _myShapeFeature.LabelXOffsetInPixels = _labelXOffsetInPixels;
                wpfMap1.Refresh(_featureOverlay);
                OnPropertyChanged("LabelXOffsetInPixels");
            }
        }

        public float LabelYOffsetInPixels
        {
            get
            {
                return _labelYOffsetInPixels;
            }
            set
            {
                _labelYOffsetInPixels = value;
                _mySymbolFeature.LabelYOffsetInPixels = _labelYOffsetInPixels;
                _myShapeFeature.LabelYOffsetInPixels = _labelYOffsetInPixels;
                wpfMap1.Refresh(_featureOverlay);
                OnPropertyChanged("LabelYOffsetInPixels");
            }
        }

        public MainWindow()
        {
            foreach (string labelLocation in Enum.GetNames(typeof(TextPlacement)))
            {
                _labelLocations.Add(labelLocation);
            }

            this.DataContext = this;
            InitializeComponent();
        }

        private void AddSimpleMarkersOverLay()
        {
            SimpleMarkerOverlay simpleMarkerOverlay = new SimpleMarkerOverlay();
            this.wpfMap1.Overlays.Add("simpleMarkerOverlay", simpleMarkerOverlay);
        }

        //private void AddSimpleMarker(PointShape point)
        //{
        //    SimpleMarkerOverlay simpleMarkerOverlay = wpfMap1.Overlays["simpleMarkerOverlay"] as SimpleMarkerOverlay;
        //    var marker = new Marker(point)
        //    {
        //        ImageSource = new BitmapImage(new Uri(System.AppDomain.CurrentDomain.BaseDirectory + "\\AQUA.png", UriKind.RelativeOrAbsolute)),
        //        Width = 1,
        //        Height = 1
        //    };
        //    //marker.MarkerMouseClick += Marker_MarkerMouseClick;

        //    // Add the marker to the simpleMarkerOverlay and refresh the map   
        //    simpleMarkerOverlay.Markers.Add(marker);
        //}

        private void ClearSimpleMarker()
        {
            SimpleMarkerOverlay simpleMarkerOverlay = wpfMap1.Overlays["simpleMarkerOverlay"] as SimpleMarkerOverlay;
            simpleMarkerOverlay.Markers.Clear();
        }


        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            try
            {
                double zoomLevel20Scale = 200.0;
                double zoomIncrement = 10.0;

                double minX = 0.0;
                double maxY = 200.0;
                double maxX = 200.0;
                double minY = 0.0;

                _gmap = wpfMap1;
                wpfMap1.MinimumScale = 0;
                wpfMap1.BackgroundOverlay.BackgroundBrush = new GeoSolidBrush(GeoColors.White);
                wpfMap1.MapUnit = GeographyUnit.Meter;
                wpfMap1.ZoomLevelSet = new MyZoomLevelSet(zoomLevel20Scale, zoomIncrement);
                wpfMap1.CurrentExtent = new RectangleShape(minX, maxY, maxX, minY);
                wpfMap1.MapTools.Logo.IsEnabled = false;
                //wpfMap1.MapTools.PanZoomBar.IsEnabled = false;

                _state0GeoImage = new GeoImage(new MemoryStream(File.ReadAllBytes("../../../App_Data/doorNormalSecure.png")));
                _state1GeoImage = new GeoImage(new MemoryStream(File.ReadAllBytes("../../../App_Data/doorNormalNotSecure.png")));
                _state2GeoImage = new GeoImage(new MemoryStream(File.ReadAllBytes("../../../App_Data/doorAlarmSecure.png")));
                _state3GeoImage = new GeoImage(new MemoryStream(File.ReadAllBytes("../../../App_Data/doorAlarmNotSecure.png")));
                _stateNoGeoImage = new GeoImage(new MemoryStream(File.ReadAllBytes("../../../App_Data/NoImageAvailable.jpg")));

                MyFeatureLayer myFeatureLayer = new MyFeatureLayer();
                myFeatureLayer.Name = $"MyFeatureLayer";

                //_mySymbolFeature = new MySymbolFeature(0, 0, 25, 25, "0");
                //myFeatureLayer.MyFeatureSource.AddMyFeature(_mySymbolFeature);

                BaseShape myFeatureShape = CreatePentagonShape(0, 0, 40, 40);
                _myShapeFeature = new MyShapeFeature(myFeatureShape, 1, 1, "0");
                myFeatureLayer.MyFeatureSource.AddMyFeature(_myShapeFeature);

                // Create the feature overlay and layers
                _featureOverlay = new LayerOverlay() { Name = "FeatureOverlay" };
                _featureOverlay.Layers.Add(myFeatureLayer);
                _featureOverlay.DrawingException += LayerOverlay_DrawingException;
                _featureOverlay.TileType = TileType.SingleTile;
                _featureOverlay.DrawingQuality = DrawingQuality.HighSpeed;
                wpfMap1.Overlays.Add(_featureOverlay);

                PointShape centerPoint = new PointShape(20, -20);
                wpfMap1.ZoomTo(centerPoint, wpfMap1.ZoomLevelSet.ZoomLevel01.Scale);

                AddSimpleMarkersOverLay();

                wpfMap1.Refresh();
            }
            catch (Exception ex)
            {
                Console.WriteLine("[Window_Loaded] Exception building map - {0}", ex.ToString());
            }
        }

        private BaseShape CreatePentagonShape(double x, double y, double width, double height)
        {
            Collection<Vertex> vertices = new Collection<Vertex>();

            vertices.Add(new Vertex(x, y));
            vertices.Add(new Vertex(x + width, y));
            vertices.Add(new Vertex(x + width, y - height));
            vertices.Add(new Vertex(x, y - height));
            vertices.Add(new Vertex(x, y));

            //vertices.Add(new Vertex(x, y + (0.75 * height)));
            //vertices.Add(new Vertex(x + (0.75 * width), y));
            //vertices.Add(new Vertex(x + (0.50 * width), y - (0.75 * height)));
            //vertices.Add(new Vertex(x - (0.50 * width), y - (0.75 * height)));
            //vertices.Add(new Vertex(x - (0.75 * width), y));

            PolygonShape polygonShape = new PolygonShape(new RingShape(vertices));

            return polygonShape;
        }

        private void wpfMap1_MouseMove(object sender, MouseEventArgs e)
        {
            //Gets the PointShape in world coordinates from screen coordinates.
            System.Windows.Point point = e.MouseDevice.GetPosition(null);

            ScreenPointF screenPointF = new ScreenPointF((float)point.X, (float)point.Y);
            PointShape pointShape = MapUtil.ToWorldCoordinate(wpfMap1.CurrentExtent, screenPointF, (float)wpfMap1.ActualWidth, (float)wpfMap1.ActualHeight);

            mouseWC.Text = "X: " + pointShape.X + "  Y: " + pointShape.Y;
            mouseSC.Text = "X: " + point.X + "  Y: " + point.Y;

            ClearSimpleMarker();
            //  wpfMap1.Refresh();

        }

        private void LayerOverlay_DrawingException(object sender, DrawingExceptionTileOverlayEventArgs e)
        {
            Console.WriteLine("[LayerOverlay_DrawingException] Layer Overlay draw exception - {0}", e.Exception.Message);
        }

        private void OnPropertyChanged(string name)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(name));
        }
    }

    public class MyFeatureLayer : FeatureLayer
    {
        public MyFeatureSource MyFeatureSource { get; private set; }

        public MyFeatureLayer()
        {
            this.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(new MyDefaultStyle());
            this.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            this.DrawingQuality = DrawingQuality.HighSpeed;
            this.DrawingMarginInPixel = 0;

            MyFeatureSource = new MyFeatureSource();
            FeatureSource = MyFeatureSource;
        }
    }

    public class MyFeatureSource : FeatureSource
    {
        public Dictionary<string, Feature> _featuresById { get; private set; } = new Dictionary<string, Feature>(StringComparer.InvariantCultureIgnoreCase);
        private Collection<Feature> _features = new Collection<Feature>();
        public override bool IsEditable { get { return true; } }

        public void AddMyFeature(MyBaseFeature feature)
        {
            if (feature != null)
            {
                _featuresById.Add(feature.Id, feature);
                _features.Add(feature);
            }
        }

        public MyBaseFeature GetMyFeature(int col, int row)
        {
            string id = $"POINT{col}.{row}";
            Feature feature = null;
            _featuresById.TryGetValue(id, out feature);

            return (MyBaseFeature)feature;
        }

        public MyBaseFeature GetMyFeature(string id)
        {
            Feature feature = null;
            _featuresById.TryGetValue(id, out feature);

            return (MyBaseFeature)feature;
        }

        protected override Collection<Feature> GetAllFeaturesCore(IEnumerable<string> returningColumnNames)
        {
            return _features;
        }

        protected override bool CanGetBoundingBoxQuicklyCore()
        {
            return true;
        }

        protected override bool CanGetCountQuicklyCore()
        {
            return true;
        }

        protected override Collection<FeatureSourceColumn> GetColumnsCore()
        {
            return base.GetColumnsCore();
        }
    }

    public abstract class MyBaseFeature : Feature
    {
        public double X { get; private set; }
        public double Y { get; private set; }
        private TextPlacement _labelLocation = TextPlacement.Lower;
        private float _labelXOffsetInPixels = 0.0F;
        private float _labelYOffsetInPixels = 0.0F;
        private TextStyle _labelStyle;
        public Collection<ThinkGeo.Core.Style> MyFeatureStyles { get; set; } = new Collection<ThinkGeo.Core.Style>();

        public const string COLUMN_ID = "Id";
        public const string COLUMN_STATE = "State";
        public const string COLUMN_LABEL = "Label";

        public WellKnownType FeatureType
        {
            get
            {
                return GetWellKnownType();
            }
        }

        public TextPlacement LabelLocation
        {
            get
            {
                return _labelLocation;
            }
            set
            {
                _labelLocation = value;
                SetLabelStyle(COLUMN_LABEL, COLUMN_LABEL);
            }
        }

        public float LabelXOffsetInPixels
        {
            get
            {
                return _labelXOffsetInPixels;
            }
            set
            {
                _labelXOffsetInPixels = value;
                SetLabelStyle(COLUMN_LABEL, COLUMN_LABEL);
            }
        }

        public float LabelYOffsetInPixels
        {
            get
            {
                return _labelYOffsetInPixels;
            }
            set
            {
                _labelYOffsetInPixels = value;
                SetLabelStyle(COLUMN_LABEL, COLUMN_LABEL);
            }
        }

        /// <summary>
        /// Gets if the user-defined X/Y label pixel offsets are to be 
        /// scaled as the map is zoomed in/out.
        /// </summary>
        protected virtual bool ScaleLabelXYPixelOffsets
        {
            get
            {
                return true;
            }
        }

        public MyBaseFeature(BaseShape shape, int col, int row, string value)
            : base(shape)
        {
            CtrInit(col, row, value);
        }

        public MyBaseFeature(int col, int row, double x, double y, string value)
            : base(x, y)
        {
            X = x;
            Y = y;

            CtrInit(col, row, value);
        }

        private void CtrInit(int col, int row, string value)
        {
            Id = $"POINT{col}.{row}";

            ColumnValues.Add(COLUMN_ID, Id);
            ColumnValues.Add(COLUMN_STATE, value);
            ColumnValues.Add(COLUMN_LABEL, Id);
            SetLabelStyle(COLUMN_LABEL, COLUMN_LABEL);
        }

        public abstract RectangleShape GetPointBoundingBox(RectangleShape currentExtent, float mapHeight, float mapWidth);

        public void SetFeatureValue(string value)
        {
            ColumnValues[COLUMN_STATE] = value;
        }

        public string GetFeatureValue()
        {
            return ColumnValues[COLUMN_STATE];
        }

        protected virtual void SetLabelStyle(string styleName, string columnName)
        {
            // Remove the label style if it already exists
            RemoveStyle(styleName);

            DrawingFontStyles dfs = DrawingFontStyles.Regular;

            System.Windows.Media.Color foreGround = System.Windows.Media.Colors.Black;
            System.Windows.Media.Color backGround = System.Windows.Media.Colors.Transparent;

            // column point label
            _labelStyle = new TextStyle(columnName,
                                                 new GeoFont("Arial", 16.0F, dfs),
                                                 new GeoSolidBrush(new GeoColor(foreGround.A, foreGround.R, foreGround.G, foreGround.B)));
            _labelStyle.Mask = new AreaStyle(new GeoSolidBrush(new GeoColor(backGround.A, backGround.R, backGround.G, backGround.B)));
            _labelStyle.Mask.DrawingLevel = DrawingLevel.LabelLevel;

            _labelStyle.DuplicateRule = LabelDuplicateRule.UnlimitedDuplicateLabels;
            _labelStyle.OverlappingRule = LabelOverlappingRule.AllowOverlapping;
            _labelStyle.TextPlacement = LabelLocation;
            _labelStyle.DrawingLevel = DrawingLevel.LabelLevel;
            _labelStyle.Name = styleName;

            _labelStyle.TextLineSegmentRatio = 9999999;//A High ratio allows the label to run past the length of the line
            this.MyFeatureStyles.Add(_labelStyle);

            // Get the bounding box of the feature in world coordinates
            RectangleShape boundingBoxWC;
            if (GetShape() is PointShape)
            {
                BaseShape baseBox = GetPointBoundingBox(MainWindow._gmap.CurrentExtent, (float)MainWindow._gmap.ActualHeight, (float)MainWindow._gmap.ActualWidth);
                boundingBoxWC = baseBox.GetBoundingBox();
            }
            else
            {
                boundingBoxWC = GetBoundingBox();
            }

            // Convert the world coordinate bounding box to screen coordinates (pixels) and 
            // position the label according to the location selected by the user
            // We need the screen coordinates because we can only alter the label location 
            // via pixels (and not world coordinates).
            // ExtentHelper.ToScreenCoordinate() sometimes throws exceptions based on the
            // ThinkGeo version if the width or height of the bounding box is < 1.0. So,
            // fix the bounding box if the width or height is < 1.0. This should guarantee
            // the width and/or height is >= 1.0.
            if (MainWindow._gmap.CurrentExtent != null)
            {
                if (boundingBoxWC.Width < 1.0)
                {
                    boundingBoxWC.LowerRightPoint.X += 1.0;
                }

                if (boundingBoxWC.Height < 1.0)
                {
                    boundingBoxWC.LowerRightPoint.Y -= 1.0;
                }

                DrawingRectangle boundingBoxSC;

                try
                {
                    boundingBoxSC = MapUtil.ToScreenCoordinate(MainWindow._gmap.CurrentExtent, boundingBoxWC, (float)MainWindow._gmap.ActualWidth, (float)MainWindow._gmap.ActualHeight);
                    if (!(GetShape() is PointShape))
                    {
                        Console.WriteLine();
                        Console.WriteLine(
                                    "***** Rectangle bounding box in SC - MinX: {0}  MaxY: {1}  MaxX: {2}  MinY: {3}  Center Point - X: {4} Y: {5}",
                                    boundingBoxSC.MinimumX,
                                    boundingBoxSC.MaximumY,
                                    boundingBoxSC.MaximumX,
                                    boundingBoxSC.MinimumY,
                                    boundingBoxSC.CenterX,
                                    boundingBoxSC.CenterY);
                        Console.WriteLine();
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Exception - {0}", ex.ToString());
                    return;
                }

                // Move the label to half the width and height of the feature so it won't 
                // be on top of the feature
                float halfWidth = boundingBoxSC.Width / 2;
                float halfHeight = boundingBoxSC.Height / 2;

                switch (LabelLocation)
                {
                    //
                    // ThinkGeo v10.3
                    //
                    // When changing the Y pixel offset a positive value "should" move the label down
                    // and a negative value "should" move the label up. In the screen coordinate system
                    // 0,0 is in the upper left corner and Y increases from the top of the screen
                    // to the bottom of the screen. In ThinkGeo v10.3 it appears that the Y screen 
                    // coordinate increases from the bottom of the screen to the top of the screen. 
                    // Hence, we have to add to Y when we really should substract from Y, and vice versa.
                    //
                    case TextPlacement.UpperLeft:
                        {
                            _labelStyle.XOffsetInPixel = -halfWidth;
                            _labelStyle.YOffsetInPixel = halfHeight;
                            break;
                        }
                    case TextPlacement.Upper:
                        {
                            _labelStyle.XOffsetInPixel = 0;
                            _labelStyle.YOffsetInPixel = halfHeight;
                            break;
                        }
                    case TextPlacement.UpperRight:
                        {
                            _labelStyle.XOffsetInPixel = halfWidth;
                            _labelStyle.YOffsetInPixel = halfHeight;
                            break;
                        }
                    case TextPlacement.Right:
                        {
                            _labelStyle.XOffsetInPixel = halfWidth;
                            _labelStyle.YOffsetInPixel = 0;
                            break;
                        }
                    case TextPlacement.Center:
                        {
                            _labelStyle.XOffsetInPixel = 0;
                            _labelStyle.YOffsetInPixel = 0;
                            break;
                        }
                    case TextPlacement.Left:
                        {
                            _labelStyle.XOffsetInPixel = -halfWidth;
                            _labelStyle.YOffsetInPixel = 0;
                            break;
                        }
                    case TextPlacement.LowerLeft:
                        {
                            _labelStyle.XOffsetInPixel = -halfWidth;
                            _labelStyle.YOffsetInPixel = -halfHeight;
                            break;
                        }
                    case TextPlacement.Lower:
                        {
                            _labelStyle.XOffsetInPixel = 0;
                            _labelStyle.YOffsetInPixel = -halfHeight;
                            break;
                        }
                    case TextPlacement.LowerRight:
                        {
                            _labelStyle.XOffsetInPixel = halfWidth;
                            _labelStyle.YOffsetInPixel = -halfHeight;
                            break;
                        }
                    default:
                        {
                            Console.WriteLine("Unknown label location '{0}'", LabelLocation);
                            _labelStyle.YOffsetInPixel = 0;
                            _labelStyle.XOffsetInPixel = 0;
                            break;
                        }


                        // ThinkGeo v12.2
                        //
                        // In v12.2 it appears the direction of Y screen coordinate has been corrected.
                        // When a positive Y offset is added the label moves down and vice versa.
                        //
                        //case TextPlacement.UpperLeft:
                        //    {
                        //        _labelStyle.XOffsetInPixel = -halfWidth;
                        //        _labelStyle.YOffsetInPixel = -halfHeight;
                        //        break;
                        //    }
                        //case TextPlacement.Upper:
                        //    {
                        //        _labelStyle.XOffsetInPixel = 0;
                        //        _labelStyle.YOffsetInPixel = -halfHeight;
                        //        break;
                        //    }
                        //case TextPlacement.UpperRight:
                        //    {
                        //        _labelStyle.XOffsetInPixel = halfWidth;
                        //        _labelStyle.YOffsetInPixel = -halfHeight;
                        //        break;
                        //    }
                        //case TextPlacement.Right:
                        //    {
                        //        _labelStyle.XOffsetInPixel = halfWidth;
                        //        _labelStyle.YOffsetInPixel = 0;
                        //        break;
                        //    }
                        //case TextPlacement.Center:
                        //    {
                        //        _labelStyle.XOffsetInPixel = 0;
                        //        _labelStyle.YOffsetInPixel = 0;
                        //        break;
                        //    }
                        //case TextPlacement.Left:
                        //    {
                        //        _labelStyle.XOffsetInPixel = -halfWidth;
                        //        _labelStyle.YOffsetInPixel = 0;
                        //        break;
                        //    }
                        //case TextPlacement.LowerLeft:
                        //    {
                        //        _labelStyle.XOffsetInPixel = -halfWidth;
                        //        _labelStyle.YOffsetInPixel = halfHeight;
                        //        break;
                        //    }
                        //case TextPlacement.Lower:
                        //    {
                        //        _labelStyle.XOffsetInPixel = 0;
                        //        _labelStyle.YOffsetInPixel = halfHeight;
                        //        break;
                        //    }
                        //case TextPlacement.LowerRight:
                        //    {
                        //        _labelStyle.XOffsetInPixel = halfWidth;
                        //        _labelStyle.YOffsetInPixel = halfHeight;
                        //        break;
                        //    }
                        //default:
                        //    {
                        //        Console.WriteLine("Unknown label location '{0}'", LabelLocation);
                        //        _labelStyle.YOffsetInPixel = 0;
                        //        _labelStyle.XOffsetInPixel = 0;
                        //        break;
                        //    }
                }

                double scaledXOffsetInPixels = LabelXOffsetInPixels;
                double scaledYOffsetInPixels = LabelYOffsetInPixels;

                // We need to scale the pixel offsets to reflect the current zoom level.
                // Not all features actually scale (e.g., icon features) so only scale 
                // the label pixel offsets if the feature scales.
                if (ScaleLabelXYPixelOffsets)
                {
                    double zoomLevel1Scale = MainWindow._gmap.ZoomLevelSet.ZoomLevel01.Scale;
                    double currentZoomLevelScale = MainWindow._gmap.CurrentScale;

                    if (currentZoomLevelScale > 0)
                    {
                        scaledXOffsetInPixels = (zoomLevel1Scale / currentZoomLevelScale) * LabelXOffsetInPixels;
                        scaledYOffsetInPixels = (zoomLevel1Scale / currentZoomLevelScale) * LabelYOffsetInPixels;
                    }
                    else
                    {
                        scaledXOffsetInPixels = LabelXOffsetInPixels;
                        scaledYOffsetInPixels = LabelYOffsetInPixels;
                    }
                }

                // Apply the specified X/Y pixel offsets defined by the user
                _labelStyle.XOffsetInPixel += (float)scaledXOffsetInPixels;
                _labelStyle.YOffsetInPixel += (float)scaledYOffsetInPixels;
            }
            else
            {
                Console.WriteLine("Map extent is null");
            }
        }

        private void RemoveStyle(string name)
        {
            ThinkGeo.Core.Style style = this.MyFeatureStyles.Where(x => x.Name == name).FirstOrDefault();
            if (style != null)
            {
                this.MyFeatureStyles.Remove(style);
            }
        }

    }

    public class MyShapeFeature : MyBaseFeature
    {
        public MyShapeFeature(BaseShape shape, int col, int row, string value)
            : base(shape, col, row, value)
        {
            MyValueStyle myValueStyle = new MyValueStyle();
            myValueStyle.ColumnName = COLUMN_STATE;
            myValueStyle.Name = "MyValueStyle";
            MyFeatureStyles.Add(myValueStyle);

            AreaStyle style = new AreaStyle(GeoPens.Black, GeoBrushes.Green);
            ValueItem vi = new ValueItem();
            vi.Value = "0";
            vi.CustomStyles.Add(style);
            myValueStyle.ValueItems.Add(vi);

            style = new AreaStyle(GeoPens.Black, GeoBrushes.Yellow);
            vi = new ValueItem();
            vi.Value = "1";
            vi.CustomStyles.Add(style);
            myValueStyle.ValueItems.Add(vi);

            style = new AreaStyle(GeoPens.Black, GeoBrushes.Red);
            vi = new ValueItem();
            vi.Value = "2";
            vi.CustomStyles.Add(style);
            myValueStyle.ValueItems.Add(vi);

            style = new AreaStyle(GeoPens.Black, GeoBrushes.Orange);
            vi = new ValueItem();
            vi.Value = "3";
            vi.CustomStyles.Add(style);
            myValueStyle.ValueItems.Add(vi);

            myValueStyle.DefaultStyle = new AreaStyle(GeoPens.Red, GeoBrushes.White);
        }

        public override RectangleShape GetPointBoundingBox(RectangleShape currentExtent, float mapHeight, float mapWidth)
        {
            return this.GetBoundingBox();
        }
    }

    public class MySymbolFeature : MyBaseFeature
    {
        public int FeatureWidthPixels { get; set; } = 100;
        public int FeatureHeightPixels { get; set; } = 100;

        /// <summary>
        /// Gets if the user-defined X/Y label pixel offsets are to be 
        /// scaled as the map is zoomed in/out.
        /// </summary>
        protected override bool ScaleLabelXYPixelOffsets
        {
            get
            {
                return false;
            }
        }

        public MySymbolFeature(int col, int row, double x, double y, string value)
            : base(col, row, x, y, value)
        {

            MyValueStyle myValueStyle = new MyValueStyle();
            myValueStyle.ColumnName = COLUMN_STATE;
            myValueStyle.Name = "MyValueStyle";
            MyFeatureStyles.Add(myValueStyle);

            PointStyle ps = new PointStyle(MainWindow._state0GeoImage);
            ValueItem vi = new ValueItem();
            vi.Value = "0";
            vi.CustomStyles.Add(ps);
            myValueStyle.ValueItems.Add(vi);

            ps = new PointStyle(MainWindow._state1GeoImage);
            vi = new ValueItem();
            vi.Value = "1";
            vi.CustomStyles.Add(ps);
            myValueStyle.ValueItems.Add(vi);

            ps = new PointStyle(MainWindow._state2GeoImage);
            vi = new ValueItem();
            vi.Value = "2";
            vi.CustomStyles.Add(ps);
            myValueStyle.ValueItems.Add(vi);

            ps = new PointStyle(MainWindow._state3GeoImage);
            vi = new ValueItem();
            vi.Value = "3";
            vi.CustomStyles.Add(ps);
            myValueStyle.ValueItems.Add(vi);

            myValueStyle.DefaultStyle = new PointStyle(MainWindow._stateNoGeoImage);
        }

        public override RectangleShape GetPointBoundingBox(RectangleShape currentExtent, float mapHeight, float mapWidth)
        {
            if (currentExtent != null)
            {
                float halfImageWidthPixels = FeatureWidthPixels / 2;
                float halfImageHeightPixels = FeatureHeightPixels / 2;

                //Get the center point of the icon in screen and world coordinates
                PointShape featureWorldCenterPoint = new PointShape(X, Y);
                ScreenPointF featureScreenCenterPoint = MapUtil.ToScreenCoordinate(currentExtent, this, mapWidth, mapHeight);

                //Get the screen coordinates of the box
                ScreenPointF upperLeftScreenPoint = new ScreenPointF(featureScreenCenterPoint.X - halfImageWidthPixels, featureScreenCenterPoint.Y - halfImageHeightPixels);
                ScreenPointF upperRightScreenPoint = new ScreenPointF(featureScreenCenterPoint.X + halfImageWidthPixels, featureScreenCenterPoint.Y - halfImageHeightPixels);
                ScreenPointF lowerRightScreenPoint = new ScreenPointF(featureScreenCenterPoint.X + halfImageWidthPixels, featureScreenCenterPoint.Y + halfImageHeightPixels);
                ScreenPointF lowerLeftScreenPoint = new ScreenPointF(featureScreenCenterPoint.X - halfImageWidthPixels, featureScreenCenterPoint.Y + halfImageHeightPixels);

                //    Console.WriteLine("[GetPointBoundingBox] {0} pixel Location '{1}'  Width '{2}' Height '{3}'", Id, featureScreenCenterPoint.ToString(), lowerRightScreenPoint.X - lowerLeftScreenPoint.X, upperRightScreenPoint.Y - lowerRightScreenPoint.Y);

                //Convert the screen coordinates of the box to world coordinates
                PointShape upperLeftWorldPoint = MapUtil.ToWorldCoordinate(currentExtent, upperLeftScreenPoint, mapWidth, mapHeight);
                PointShape upperRightWorldPoint = MapUtil.ToWorldCoordinate(currentExtent, upperRightScreenPoint, mapWidth, mapHeight);
                PointShape lowerRightWorldPoint = MapUtil.ToWorldCoordinate(currentExtent, lowerRightScreenPoint, mapWidth, mapHeight);
                PointShape lowerLeftWorldPoint = MapUtil.ToWorldCoordinate(currentExtent, lowerLeftScreenPoint, mapWidth, mapHeight);

                //    Console.WriteLine("[GetPointBoundingBox] {0} map Location '{1}'  Width '{2}' Height '{3}'", Id, featureWorldCenterPoint.ToString(), lowerRightWorldPoint.X - lowerLeftWorldPoint.X, upperRightWorldPoint.Y - lowerRightWorldPoint.Y);

                //Create the bounding box using the calculated 4 points as vertices
                Collection<Vertex> vertices = new Collection<Vertex>();
                vertices.Add(new Vertex(upperLeftWorldPoint.X, upperLeftWorldPoint.Y));
                vertices.Add(new Vertex(upperRightWorldPoint.X, upperRightWorldPoint.Y));
                vertices.Add(new Vertex(lowerRightWorldPoint.X, lowerRightWorldPoint.Y));
                vertices.Add(new Vertex(lowerLeftWorldPoint.X, lowerLeftWorldPoint.Y));
                PolygonShape rectangleShape = new PolygonShape(new RingShape(vertices));

                //RectangleShape box = rectangleShape.GetBoundingBox();
                //if (MainWindow._logDebug)
                //    Console.WriteLine("[GetPointBoundingBox] {0} map bounding box '{1}'  Width '{2}'  Height '{3}'", Id, box.ToString(), box.LowerRightPoint.X - box.LowerLeftPoint.X, box.UpperRightPoint.Y - box.LowerRightPoint.Y);

                return rectangleShape.GetBoundingBox();
            }
            else
            {
                return GetBoundingBox();
            }
        }
    }

    public class MyValueStyle : ValueStyle
    {
        public ThinkGeo.Core.Style DefaultStyle { get; set; }

        protected override void DrawCore(IEnumerable<Feature> features, GeoCanvas canvas, Collection<SimpleCandidate> labelsInThisLayer, Collection<SimpleCandidate> labelsInAllLayers)
        {
            //    Console.WriteLine("[MyValueStyle.DrawCore  ] Draw '{0,2}' features", features.Count());

            foreach (Feature feature in features)
            {
                MyBaseFeature myFeature = (MyBaseFeature)feature;
                string fieldValue = feature.ColumnValues[ColumnName].Trim();
                ValueItem valueItem = GetValueItem(fieldValue);

                Feature[] tmpFeatures = new Feature[1] { feature };

                if (valueItem.CustomStyles.Count == 0)
                {
                    if (!string.IsNullOrEmpty(valueItem.Value))
                    {
                        valueItem.DefaultAreaStyle.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                        valueItem.DefaultLineStyle.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                        valueItem.DefaultPointStyle.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                    }
                    else
                    {
                        DefaultStyle.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                    }

                    valueItem.DefaultTextStyle.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                }
                else
                {
                    foreach (ThinkGeo.Core.Style style in valueItem.CustomStyles)
                    {
                        style.Draw(tmpFeatures, canvas, labelsInThisLayer, labelsInAllLayers);
                    }
                }
            }
        }

        private ValueItem GetValueItem(string columnValue)
        {
            ValueItem valueItem = null;

            foreach (ValueItem item in ValueItems)
            {
                if (string.Compare(columnValue, item.Value, StringComparison.OrdinalIgnoreCase) == 0)
                {
                    valueItem = item;
                    break;
                }
            }

            if (valueItem == null)
            {
                valueItem = new ValueItem();
            }

            return valueItem;
        }

    }

    public class MyDefaultStyle : ThinkGeo.Core.Style
    {
        protected override void DrawCore(IEnumerable<Feature> features, GeoCanvas canvas, Collection<SimpleCandidate> labelsInThisLayer, Collection<SimpleCandidate> labelsInAllLayers)
        {
            Console.WriteLine("[MyDefaultStyle.DrawCore] {0} Draw '{1,2}' features", DateTime.Now.ToLongTimeString(), features.Count());

            try
            {
                Stopwatch sw = new Stopwatch();
                sw.Start();

                foreach (Feature feature in features)
                {
                    MyBaseFeature myBaseFeature = feature as MyBaseFeature;
                    if (myBaseFeature != null)
                    {
                        //Draw all the styles used by this feature
                        foreach (ThinkGeo.Core.Style style in myBaseFeature.MyFeatureStyles)
                        {
                            try
                            {
                                style.Draw(new Collection<Feature>() { feature }, canvas, labelsInThisLayer, labelsInAllLayers);
                            }
                            catch (Exception ex)
                            {
                                Console.WriteLine("[MyDefaultStyle.DrawCore] Exception drawing styles for point '{0}' - {1}", myBaseFeature.Id, ex.ToString());
                            }
                        }
                    }
                }

                sw.Stop();
                Console.WriteLine("[MyDefaultStyle.DrawCore] Draw features time - {0}", sw.Elapsed.ToString());
            }
            catch (Exception ex)
            {
                Console.WriteLine("[MyDefaultStyle.DrawCore] Exception drawing features - {0}", ex.ToString());
            }
        }
    }

    public class MyZoomLevelSet : ZoomLevelSet
    {
        public MyZoomLevelSet(double start, double offset)
        {
            ZoomLevel20.Scale = start;
            ZoomLevel19.Scale = start += offset;
            ZoomLevel18.Scale = start += offset;
            ZoomLevel17.Scale = start += offset;
            ZoomLevel16.Scale = start += offset;
            ZoomLevel15.Scale = start += offset;
            ZoomLevel14.Scale = start += offset;
            ZoomLevel13.Scale = start += offset;
            ZoomLevel12.Scale = start += offset;
            ZoomLevel11.Scale = start += offset;
            ZoomLevel10.Scale = start += offset;
            ZoomLevel09.Scale = start += offset;
            ZoomLevel08.Scale = start += offset;
            ZoomLevel07.Scale = start += offset;
            ZoomLevel06.Scale = start += offset;
            ZoomLevel05.Scale = start += offset;
            ZoomLevel04.Scale = start += offset;
            ZoomLevel03.Scale = start += offset;
            ZoomLevel02.Scale = start += offset;
            ZoomLevel01.Scale = start += offset;

            Console.WriteLine("[MyZoomLevelSet] Zoom level 20 scale is '{0}'", ZoomLevel20.Scale);
            Console.WriteLine("[MyZoomLevelSet] Zoom level 1 scale is '{0}'", ZoomLevel01.Scale);
        }
    }
}
